/*
 * Copyright (c) 1996 University College London
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *      This product includes software developed by the Computer Science
 *      Department at University College London
 * 4. Neither the name of the University nor of the Department may be used
 *    to endorse or promote products derived from this software without
 *    specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE UNIVERSITY AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE UNIVERSITY OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 */
#include <string.h>
#include <search.h>
#include "prototypes.h"
#define PARANOIA
/*#define DEBUG*/
/*#define ABORT*/



void get_last_mod_time(p, t)
     page *p;
     struct timeval *t;
{
  time_copy(&(p->last_mod), t);
}

block *find_block_by_id(p, blockid)
     page *p;
     char *blockid;
{
  block *b;
  static char cacheblockid[ID_LEN];
  static block *cacheblock=NULL;
#ifdef PARANOIA
  ENTRY *item, searchitem;
#endif
  /*keep a one block cache to speed this up somewhat*/
  if (strncmp(cacheblockid, blockid, ID_LEN)==0)
    {
      return cacheblock;
    }


  if (p->first_block==NULL)
    return NULL;
  else
    b=p->first_block;
  while(strncmp(b->blockid, blockid, ID_LEN)!=0)
    {
      if(b->next_block==NULL)
	{
#ifdef DEBUG
	  printf("no block called >%s<\n", blockid);
#ifdef ABORT
	  if (strcmp(blockid,"NULL")==0) abort();
#endif
#endif
	  return NULL;
	}
      else
	b=b->next_block;
    }
#ifdef PARANOIA
  searchitem.key=blockid;
  item=hsearch(searchitem, FIND);
  if((item!=NULL) && ((block*)(item->data)!=(block*)b))
    {
      if(strcmp(((block *)(item->data))->blockid, b->blockid)==0)
	{
	  fprintf(stderr, "Paranoia check in ds_fns.c - got a duplicated block\n %s != %s\n", ((block *)(item->data))->blockid, b->blockid);
	  abort();
	}
      else
	{
	  printf("Hash failed with block id >%s<\n", blockid);
#ifdef ABORT
	  abort();
#endif
	}
    }
#endif
  cacheblock=b;
  strncpy(cacheblockid, blockid, ID_LEN);
  return b;
}

line *find_line_by_id(lineid, b)
     char *lineid;
     block *b;
{
  line *l;
  static char cachelineid[ID_LEN];
  static line *cacheline=NULL;

  /*keep a one line cache to speed up repeated accesses*/  
  if(strncmp(cachelineid, lineid, ID_LEN)==0)
    {
#ifdef NOTDEF
      printf("find_line_by_id returning cache entry %x\n", cacheline);
#endif
      return cacheline;
    }

  if (b==NULL)
    {
      printf("Trying to find line in null block\n");
      return NULL;
    }
  if (b->first_line==NULL)
    return NULL;
  else
    l=b->first_line;
  while(strncmp(l->lineid, lineid, ID_LEN)!=0)
    {
      if(l->next_line==NULL)
	return NULL;
      else
	l=l->next_line;
    }
  cacheline=l;
  strncpy(cachelineid, lineid, ID_LEN);
  return l;
}
  

line *find_last_line(b)
     block *b;
{
  line *l;
  if (b==NULL)
    {
      printf("Trying to find line in null block\n");
      return NULL;
    }
  if (b->first_line==NULL)
    return NULL;
  else
    l=b->first_line;
  while((l->next_line)!=NULL)
    {
      l=l->next_line;
    }
  return l;
}
  
int count_lines(block *b)
{
  line *l;
  int lines=0;
  if (b==NULL)
    {
      printf("Trying to find line in null block\n");
      return 0;
    }
  if (b->first_line==NULL)
    return 0;
  else
    {
      l=b->first_line;
    }
  while(l!=NULL)
    {
      if(l->linenum != LINE_DELETED) lines++;
      l=l->next_line;
    }
  return lines;
}
  
void find_best_place_for_line(block *b, u_int16 lnum, line **prevlp, 
			      line **nextlp)
{
  line *l;
  if (b==NULL)
    {
      printf("Trying to find line in null block\n");
      *prevlp=NULL;
      *nextlp=NULL;
      return;
    }
  if (b->first_line==NULL)
    {
#ifdef DEBUG
      printf("No previous lines\n");
#endif
      *prevlp=NULL;
      *nextlp=NULL;
      return;
    }
  else
    l=b->first_line;
  while(1)
    {
      if((l->linenum!=LOCAL_LINE)&(l->linenum>lnum))
	{
	  /*we've gone one past where we wanted to be*/
	  if(l->prev_line==NULL)
	    {
	      *prevlp=NULL;
#ifdef DEBUG
	      printf("best place for line is start\n");
#endif
	    }
	  else
	    {
#ifdef DEBUG
	      printf("best place for line is after line %x\n", l->lineid);
#endif
	      *prevlp=l->prev_line;
	    }
	  *nextlp=l;
	  return;
	}
      if(l->next_line!=NULL)
	l=l->next_line;
      else
	{
#ifdef DEBUG
	  printf("best place for line is end\n");
#endif
	  *prevlp=l;
	  *nextlp=NULL;
	  return;
	}
    }
}
  
void delete_line(l, b)
     line *l;
     block *b;
{
  /*this doesn't actually delete anything - it just moves the line to
    the end of the block's list to get it out of the way*/
  line *last, *prev, *next;
  if(l==NULL) return;
  if(b==NULL) return;
  last=find_last_line(b);
  if(last==l)
    {
      b->no_of_lines--;
      return;
    }
  prev=l->prev_line;
  next=l->next_line;
  last->next_line=l;
  l->prev_line=last;
  l->next_line=NULL;
  if(prev!=NULL) prev->next_line=next;
  if(next!=NULL) next->prev_line=prev;
  b->no_of_lines--;
}

int print_blocks(page *p)
{
  block *b;
  b=p->first_block;
  while(b!=NULL)
    {
      fprintf(stderr, "%s: %s\n     %s\n", b->blockid, b->originator.username,
	      b->first_line->line_data);
      b=b->next_block;
    }
  return 0;
}

